#include "goodmanwearehistogrambuilder.h"
#include "function.h"
#include "randomnumbergenerator.h"
#include <stdlib.h>
#include <iostream>
#include <cmath>

using namespace meanwalker;

// WARNING: This is an insanely BAD random number generator. For real
//          applications, use a decent one (e.g. the Mersenne twister,
//          available in the GNU Scientific Library)

class MyRandomNumberGenerator : public RandomNumberGenerator
{
public:
	MyRandomNumberGenerator()	{ }
	~MyRandomNumberGenerator()	{ }

	double pickRandomNumber()
	{
		double randomNumber = (1.0*((double)rand()/(RAND_MAX + 1.0)));

		return randomNumber;
	}
};

// This is the logarithm of the probability density we're going to
// sample: a gaussian distribution with specific mean and standard
// deviation

class LogGaussianFunction : public Function
{
public:
	LogGaussianFunction(double center, double sigma)
	{
		m_center = center;
		m_sigma = sigma;
	}

	size_t getDimension() const
	{
		return 1;
	}

	double operator()(const double *pCoords) const
	{
		double x = pCoords[0];
		double diff = (x-m_center)/m_sigma;
		return -diff*diff/2.0;
	}

	bool isLog() const
	{
		return true;
	}
private:
	double m_center, m_sigma;
};

// Our class derived from GoodmanWeareHistogramBuilder reimplements the
// getHistogramMinMax method. This is the function that's used to determine
// the histogram boundaries when the values are recorded at first, and
// processed into a histogram later.

class MyBuilder : public GoodmanWeareHistogramBuilder
{
public:
	MyBuilder(RandomNumberGenerator *pRng) : GoodmanWeareHistogramBuilder(pRng)
	{
	}
protected:
	void getHistogramMinMax(size_t idx, const double *pValues, size_t numValues, double &minValue, double &maxValue)
	{
		double mean = 0;

		for (size_t s = 0 ; s < numValues ; s++)
			mean += pValues[s];

		mean /= (double)numValues;

		double stdDev = 0;

		for (size_t s = 0 ; s < numValues ; s++)
		{
			double diff = pValues[s] - mean;

			stdDev += diff*diff;
		}

		stdDev /= (double)numValues;
		stdDev = std::sqrt(stdDev);

		minValue = mean - 2.0*stdDev;
		maxValue = mean + 2.0*stdDev;
	}
};

int main(void)
{
	MyRandomNumberGenerator rndGen;
	GoodmanWeareHistogramBuilder builder1(&rndGen);
	LogGaussianFunction f(10, 3);

	double startCoord[] = { -20000 };
	size_t numBins = 100;

	// First, we'll use a default histogram builder based on
	// the Goodman-Weare method. This method has the advantage
	// that no step scale needs to be set (e.g. by specifying
	// minimum and maximum coordinates like in the Metropolis-Hastings
	// based builder). 
	//
	// We'll record the values first, and process them into
	// a histogram later. The default procedure to calculate the
	// histogram boundaries just uses the minimum and maximum values
	// recorded.
	//
	// Note that we're sampling a gaussian distribution centered on
	// 10, and we let the walkers start around -20000, obviously
	// very far away. This Goodman-Weare method has no problem in
	// automatically sampling the true density even under these
	// circumstances.

	if (!builder1.init(&f, startCoord, 1.0, numBins))
	{
		std::cerr << builder1.getErrorString() << std::endl;
		return -1;
	}

	builder1.run(10000, 1000000);
	builder1.printHistograms();

	// In the second case, a derived class is used which reimplements
	// the method that estimates the histogram boundaries from the
	// recorded data. The estimate used is obviously not that good as
	// we don't get to see some interesting parts of the probability
	// distribution.

	MyBuilder builder2(&rndGen);

	if (!builder2.init(&f, startCoord, 1.0, numBins))
	{
		std::cerr << builder2.getErrorString() << std::endl;
		return -1;
	}

	builder2.run(10000, 1000000);
	builder2.printHistograms();

	return 0;
}
