#include "goodmanwearehistogrambuilder.h"
#include "function.h"
#include "randomnumbergenerator.h"
#include "histogram2d.h"
#include <stdlib.h>
#include <iostream>
#include <cmath>

using namespace meanwalker;

// WARNING: This is an insanely BAD random number generator. For real
//          applications, use a decent one (e.g. the Mersenne twister,
//          available in the GNU Scientific Library)

class MyRandomNumberGenerator : public RandomNumberGenerator
{
public:
	MyRandomNumberGenerator()	{ }
	~MyRandomNumberGenerator()	{ }

	double pickRandomNumber()
	{
		double randomNumber = (1.0*((double)rand()/(RAND_MAX + 1.0)));

		return randomNumber;
	}
};

// This specifies the logarithm of a rotated 2D gaussian distribution,
// in which one axis is much smaller than the other. This is one of the
// examples from the introduction from the paper by Goodman and Weare.

class LogGaussianFunction2D : public Function
{
public:
	LogGaussianFunction2D(double epsilon)
	{
		m_epsilon = epsilon;
	}

	size_t getDimension() const
	{
		return 2;
	}

	double operator()(const double *pCoords) const
	{
		double x1 = pCoords[0];
		double x2 = pCoords[1];

		double diff = x1-x2;
		double sum = x1+x2;

		return -diff*diff/(2.0*m_epsilon) - sum*sum/2.0;
	}

	bool isLog() const
	{
		return true;
	}
private:
	double m_epsilon;
};

// We'll derive our own builder from the Goodman-Weare based histogram builder
// and reimplement the onNewSample function. This allows us to create a 2D
// histogram as well.

class MyBuilder : public GoodmanWeareHistogramBuilder
{
public:
	MyBuilder(RandomNumberGenerator *pRng) : GoodmanWeareHistogramBuilder(pRng)
	{
		m_hist2D.init(-3, 3, 50, -3, 3, 50);
	}

	void printHistogram2D()
	{
		m_hist2D.print(false);
	}

protected:
	void onNewSample(double functionValue, const double *pCoords)
	{
		// We want to preserve the functions from the base class,
		// so we have to allow it to process the sample as well.
		GoodmanWeareHistogramBuilder::onNewSample(functionValue, pCoords);

		m_hist2D.process(pCoords[0], pCoords[1]);
	}

	Histogram2D m_hist2D;
};

int main(void)
{
	MyRandomNumberGenerator rndGen;
	MyBuilder builder(&rndGen);
	LogGaussianFunction2D f(0.01);

	double startCoord[] = { 0, 0 };
	size_t numBins = 100;

	if (!builder.init(&f, startCoord, 1, numBins))
	{
		std::cerr << builder.getErrorString() << std::endl;
		return -1;
	}

	builder.run(10000, 1000000);
	builder.printHistograms();
	builder.printHistogram2D();

	return 0;
}
