/*

  This file is a part of MeanWalker, a library which provides utilities to
  sample from probability distributions using methods like the 
  Metropilis-Hastings algorithm and Goodman-Weare algorithm.

  Copyright (C) 2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "histogram.h"
#include <iostream>

namespace meanwalker
{

Histogram::Histogram()
{
	m_maxValue = 0;
	m_minValue = 0;
	m_totalCount = 0;
	m_outsideCount = 0;
	m_binSize = 1;
	m_doubleBins = 0;
}

Histogram::~Histogram()
{
}

bool Histogram::init(double minValue, double maxValue, size_t numBins)
{
	if (m_histogram.size() != 0)
	{
		setErrorString("Already initialized");
		return false;
	}

	if (numBins < 1)
	{
		setErrorString("There should be at least one bin");
		return false;
	}

	m_histogram.resize(numBins);
	for (size_t i = 0 ; i < numBins ; i++)
		m_histogram[i] = 0;

	m_minValue = minValue;
	m_maxValue = maxValue;
	m_binSize = (maxValue-minValue)/(double)numBins;
	m_doubleBins = (double)numBins;
	m_totalCount = 0;
	m_outsideCount = 0;
	return true;
}

void Histogram::clear()
{
	size_t numBins = m_histogram.size();

	for (size_t i = 0 ; i < numBins ; i++)
		m_histogram[i] = 0;

	m_totalCount = 0;
	m_outsideCount = 0;
}

void Histogram::destroy()
{
	m_histogram.resize(0);
	m_maxValue = 0;
	m_minValue = 0;
	m_totalCount = 0;
	m_outsideCount = 0;
	m_binSize = 1;
	m_doubleBins = 0;
}

void Histogram::print() const
{
	int numBins = m_histogram.size();

	if (numBins == 0)
		return;

	for (int j = 0 ; j < numBins ; j++)
	{
		std::cout << (getBinCenter(j)-m_binSize/2.0) << "\t" << getProbabilityDensity(j) << std::endl;
		std::cout << (getBinCenter(j)+m_binSize/2.0) << "\t" << getProbabilityDensity(j) << std::endl;
	}
	std::cout << std::endl;
	std::cout << std::endl;
}

DelayedHistogram::DelayedHistogram()
{
	m_recPos = 0;
}

bool DelayedHistogram::processRecordedValues(size_t numBins)
{
	if (m_recordedValues.empty() || m_recPos == 0)
	{
		setErrorString("No recorded values present");
		return false;
	}
	
	double minValue, maxValue;

	getHistogramMinMax(&(m_recordedValues[0]), m_recPos, minValue, maxValue);

	if (!init(minValue, maxValue, numBins))
		return false;

	for (size_t i = 0 ; i < m_recPos ; i++)
		process(m_recordedValues[i]);

	return true;
}

void DelayedHistogram::getDefaultHistogramMinMax(const double *pValues, size_t numValues, double &minValue, double &maxValue)
{
	if (numValues == 0)
	{
		minValue = 0;
		maxValue = 1;
	}

	double minV = pValues[0];
	double maxV = pValues[0];

	for (size_t i = 1 ; i < numValues ; i++)
	{
		double x = pValues[i];

		if (x > maxV)
			maxV = x;
		if (x < minV)
			minV = x;
	}

	if (minV == maxV)
	{
		double frac = minV*1e-10;

		if (frac < 0)
			frac = -frac;

		if (frac == 0)
			frac = 1e-10;

		minV -= frac;
		maxV += frac;
	}

	minValue = minV;
	maxValue = maxV;
}

} // end namespace   

