/*

  This file is a part of MeanWalker, a library which provides utilities to
  sample from probability distributions using methods like the 
  Metropilis-Hastings algorithm and Goodman-Weare algorithm.

  Copyright (C) 2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "metropoliswalker.h"
#include "randomnumbergenerator.h"
#include <cmath>

#define METROPOLISWALKER_PI 3.14159265358979323846264338327950288419716939937510582097494459230

namespace meanwalker
{

MetropolisWalker::MetropolisWalker(RandomNumberGenerator *pRndGen) : Walker(pRndGen)
{
}

MetropolisWalker::~MetropolisWalker()
{
}

bool MetropolisWalker::init(Function *pProbFunction, const double *pMinCoords, const double *pMaxCoords, const double *pStartCoords, bool keepInside)
{
	if (getProbabilityFunction() != 0)
	{
		setErrorString("Already initialized");
		return false;
	}

	if (pProbFunction == 0)
	{
		setErrorString("You must specify a probability function");
		return false;
	}

	setProbabilityFunction(pProbFunction);
	
	size_t dim = pProbFunction->getDimension();
	RandomNumberGenerator *pRndGen = getRandomNumberGenerator();

	m_minCoords.resize(dim);
	m_maxCoords.resize(dim);
	m_coords.resize(dim);
	m_bestCoords.resize(dim);

	for (size_t i = 0 ; i < dim ; i++)
	{
		if (pMaxCoords[i] <= pMinCoords[i])
		{
			setErrorString("Maximum coordinates should all be larger than minimum ones");
			return false;
		}
		m_minCoords[i] = pMinCoords[i];
		m_maxCoords[i] = pMaxCoords[i];

		if (pStartCoords == 0)
		{
			double x = pRndGen->pickRandomNumber();

			m_coords[i] = (m_maxCoords[i] - m_minCoords[i])*x + m_minCoords[i];
		}
		else
			m_coords[i] = pStartCoords[i];
		m_bestCoords[i] = m_coords[i];
	}
	m_tryCoords.resize(dim);

	m_stepScale.resize(dim);
	for (size_t i = 0 ; i < dim ; i++)
		m_stepScale[i] = (pMaxCoords[i]-pMinCoords[i])/1000.0;

	m_keepInside = keepInside;

	return true;
}

void MetropolisWalker::setNewStartCoordinates(double *pStartCoords)
{
	size_t dim = getDimension();
	RandomNumberGenerator *pRndGen = getRandomNumberGenerator();

	for (size_t i = 0 ; i < dim ; i++)
	{
		if (pStartCoords == 0)
		{
			double x = pRndGen->pickRandomNumber();

			m_coords[i] = (m_maxCoords[i] - m_minCoords[i])*x + m_minCoords[i];
		}
		else
			m_coords[i] = pStartCoords[i];
	}
}

void MetropolisWalker::walk(size_t numSteps)
{
	Function &probFunction = *(getProbabilityFunction());
	double curValue = probFunction(&(m_coords[0]));
	double bestValue = probFunction(&(m_bestCoords[0]));
	size_t dim = m_coords.size();
	RandomNumberGenerator *pRndGen = getRandomNumberGenerator();
	bool isLog = probFunction.isLog();

	for (size_t i = 0 ; i < numSteps ; i++)
	{
		stepToNewCoords(&(m_coords[0]), &(m_tryCoords[0]));

		bool inside = true;

		if (m_keepInside)
		{
			for (size_t j = 0 ; inside && j < dim ; j++)
			{
				if (m_tryCoords[j] < m_minCoords[j])
					inside = false;
				if (m_tryCoords[j] > m_maxCoords[j])
					inside = false;
			}
		}

		if (!probFunction.isValid(&(m_tryCoords[0])))
			inside = false;

		bool accept = false;
		double newValue = 0;

		if (inside)
		{
			newValue = probFunction(&(m_tryCoords[0]));

			if (newValue > curValue)
				accept = true;
			else
			{
				double frac = 0;

				if (isLog)
					frac = std::exp(newValue-curValue);
				else
					frac = newValue/curValue;

				double randomValue = pRndGen->pickRandomNumber();
				
				if (randomValue < frac)
					accept = true;
			}
		}

		if (accept)
		{
			for (size_t j = 0 ; j < dim ; j++)
			{
				m_coords[j] = m_tryCoords[j];
				curValue = newValue;
			}
		}

		if (curValue > bestValue)
		{
			bestValue = curValue;
			for (size_t j = 0 ; j < dim ; j++)
				m_bestCoords[j] = m_coords[j];
		}

		onNewSample(curValue, &(m_coords[0]));
	}
}

void MetropolisWalker::stepToNewCoords(const double *pCurCoords, double *pNewCoords)
{
	size_t dim = getDimension();
	RandomNumberGenerator *pRndGen = getRandomNumberGenerator();

	// TODO: this is a lorentzian distribution, try others?

	for (size_t i = 0 ; i < dim ; i++)
	{
		double z = pRndGen->pickRandomNumber();
		double x = -m_stepScale[i]/std::tan(z*METROPOLISWALKER_PI);

		pNewCoords[i] = pCurCoords[i] + x;
	}
}

} // end namespace   
